(* Comments in Rocq go between "(*" and "*)". *)

(* comments (* can (* be (* nested *)*)*)*)


Module Booleans.

(*
Inductive bool :=
| true
| false.
*)

Inductive bool : Set :=
| true : bool
| false : bool.

Check bool.
Check Set.
Check Type.

Definition negb (b : bool) : bool :=
    match b with
    | true => false
    | false => true
    end.

Compute negb true.
Compute negb false.

Definition andb (b1 : bool) (b2 : bool) : bool :=
    match b1 with
    | true => b2
    | false => false
    end.
Check andb.

Compute andb true true.
Compute andb true false.
Compute andb false true.
Compute andb false false.

Lemma andb_comm :
  forall b1 b2,
    andb b1 b2 = andb b2 b1.
Proof.
  intro b1.
  intro b2.
  destruct b1.
  - simpl.
    destruct b2.
    + simpl.
      reflexivity.
    + auto.
  - simpl.
    destruct b2.
    + simpl.
      reflexivity.
    + auto.
Qed.

Lemma andb_comm' :
  forall b1 b2,
    andb b1 b2 = andb b2 b1.
Proof.
  destruct b1; destruct b2; auto.
Qed.

Print andb_comm. 



(* currying and function types *)
Check negb.
Check andb.

Definition andb_no_sugar : bool -> bool -> bool :=
  fun b1 =>
    fun b2 =>
      match b1 with
      | true => b2
      | false => false
      end.

Print andb.
Print andb_no_sugar.

End Booleans.


Module Nats.

Inductive nat : Set :=
| O : nat
| S : nat -> nat.

Check O.           (* 0 *)
Check S O.         (* 1 *)
Check S (S O).     (* 2 *)
Check S (S (S O)). (* 3 *)

Definition isZero (n : nat) : bool :=
  match n with
  | O => true
  | S _ => false
  end.

Compute isZero O.         (* true  *)
Compute isZero (S O).     (* false *)
Compute isZero (S (S O)). (* false *)

Definition isOne (n : nat) : bool :=
  match n with
  | O   => false
  | S O => true
  | _   => false
  end.

Compute isOne O.         (* false *)
Compute isOne (S O).     (* true  *)
Compute isOne (S (S O)). (* false *)

Fail Definition add (n1 : nat) (n2 : nat) : nat :=
  match n1 with
  | O => n2
  (* | S m1 => add m1 (S n2) *)
  | S m1 => S (add m1 n2)
  end.

Fixpoint add (n1 : nat) (n2 : nat) : nat :=
  match n1 with
  | O => n2
  | S foo => S (add foo n2)
  end.

Fixpoint add' (n1 : nat) (n2 : nat) : nat :=
  match n1 with
  | O => n2
  | S m1 => add' m1 (S n2)
  end.

Lemma the_thing_we_need :
  forall n1 n2,
    S (add' n1 n2) = add' n1 (S n2).
Proof.
  intro n1.
  (* intro n2. *)
  induction n1.
  - simpl. reflexivity.
  - simpl. intro n2.
    rewrite IHn1.
    reflexivity.
Qed.

Theorem add_add'_are_the_same_thing_according_to_james : 
  forall n1 n2,
    add n1 n2 = add' n1 n2.
Proof.
  intro n1.
  intro n2.
  induction n1.
  - simpl. reflexivity.
  - simpl. rewrite IHn1.
    rewrite the_thing_we_need.
    reflexivity.
Qed.

Print add.

Lemma add_O_n :
  forall n,
    add O n = n.
Proof.
  auto.
Qed.

Lemma add_n_O :
  forall n,
    add n O = n.
Proof.
  induction n; auto.
  simpl; f_equal.
  assumption.

  (*
  intro n.
  destruct n as [|m].
  - simpl. reflexivity.
  - simpl. f_equal.
    destruct m as [|l].
    + auto.
    + simpl. f_equal.
      destruct l as [|k].
      * auto.
      * simpl; f_equal.
        destruct k as [|j]; auto; simpl; f_equal. 
        destruct j as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 
        destruct n as [|n]; auto; simpl; f_equal. 

Admitted.
  *)
Qed.

Fixpoint mul (n1 : nat) (n2 : nat) : nat :=
  match n1 with
  | O => O
  | S m1 => add n2 (mul m1 n2)
  end.

Compute mul (S (S O)) (S (S O)). (* S (S (S (S O))) *)

Fixpoint factorial (n : nat) : nat :=
  match n with
  | O => S O
  | S m => mul n (factorial m)
  end.

Compute factorial (S (S (S O))). (* S (S (S (S (S (S O))))) *)

Fixpoint factorial_tr (n : nat) (acc : nat) : nat :=
  match n with
  | O => acc
  | S m => factorial_tr m (mul n acc)
  end.

Definition factorial_tailrec (n : nat) : nat :=
  factorial_tr n (S O).

End Nats.

Check nat.
Print nat.

Module Options.


Check 3.
(* Set Printing All. *)
Check 3.


Inductive maybe_nat :=
| WeHaveANat : nat -> maybe_nat
| WeDontHaveANat : maybe_nat.

(* return the previous natural number *)
Definition pred_by_hand (n : nat) : maybe_nat :=
  match n with
  | O => WeDontHaveANat
  | S m => WeHaveANat m
  end.

  Inductive maybe_bool : Set :=
  | WeHaveABool : bool -> maybe_bool
  | WeDontHaveABool : maybe_bool.


Inductive option (T : Type) : Type :=
| None : option T
| Some : T -> option T.

Check option.
Check None.
Check Some.

Check Some nat 3.
(* Some<nat>(3) *)
Check Some _ 3.










Check option. (* option : Type -> Type *)
Check None.   (* forall (T : Type), option T *)
Check Some.   (* forall (T : Type), T -> option T *)

Check None bool. (* None bool : option bool *)
Check None nat.  (* None nat  : option nat  *)

Check Some bool true.  (* Some bool true : option bool *)
Check Some nat  (S O). (* Some nat (S O) : option nat  *)

(* forall is just "->" generalized! *)
Check Nat.add : forall (n1 : nat), forall (n2 : nat), nat.

(* we can also ask for type inference *)
Arguments None {T}.
Arguments Some {T} _.





Check None. (* None : option ?T *)
Check Some. (* Some : ?T -> option ?T *)

Check Some true.  (* Some true : option bool *)
Check Some (S O). (* Some (S O) : option nat  *)

Definition pred (n : nat) : option nat :=
  match n with
  | O => None
  | S m => Some m
  end.

Compute pred 0. (* None   *)
Compute pred 1. (* Some 0 *)
Compute pred 2. (* Some 1 *)

Lemma pred_None_inv :
  forall n,
    pred n = None ->
    n = 0.
Proof.
  intro n.
  intro H.
  destruct n as [|p].
  - reflexivity.
  - simpl in H.
    (* inversion H. *)
    congruence.
Qed.
Print Assumptions pred_None_inv.

Print pred_None_inv.

End Options.





Require Import Arith.

Fixpoint factorial (n : nat) : nat :=
  match n with
  | O => 1
  | S m => n * factorial m
  end.

(*
factorial 0 = 1
factorial n = n * factorial (n - 1)   if n > 0
*)

Compute factorial (S (S (S O))). (* S (S (S (S (S (S O))))) *)
Compute factorial 4.

Fixpoint factorial_tr (n : nat) (acc : nat) : nat :=
  match n with
  | O => acc
  | S m => factorial_tr m (n * acc)
  end.

(*
acc = 10
while n > 0:
  acc = acc * n
  n = n - 1

*)

Definition factorial_tailrec (n : nat) : nat :=
  factorial_tr n 1.

Compute factorial_tailrec 4.

Require Import Lia.

Lemma zachs_lemma :
  forall n acc,
    factorial_tr n acc = acc * factorial n.
Proof.
  intros n.
  induction n.
  - simpl. lia.
  - intros. simpl. rewrite IHn. lia.
Qed.

Definition factorial_equivalent_factorial_tailrec :
  forall n,
    factorial n = factorial_tailrec n.
Proof.
  intro n.
  unfold factorial_tailrec.
  rewrite zachs_lemma.
  simpl.
  lia.
Qed.

Print Assumptions factorial_equivalent_factorial_tailrec.


  induction n.
  - reflexivity.
  - simpl.
    unfold factorial_tailrec.
    simpl.
    unfold factorial_tailrec in IHn.
    rewrite IHn.
